/* modprobe.c: insert a module into the kernel, intelligently.
    Copyright (C) 2001  Rusty Russell.
    Copyright (C) 2002  Rusty Russell, IBM Corporation.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <sys/utsname.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <dirent.h>
#include <limits.h>
#include <elf.h>
#include <getopt.h>
#include <fnmatch.h>
#include <asm/unistd.h>
#include <sys/wait.h>

#include "list.h"
#include "backwards_compat.c"

struct module {
	struct list_head list;
	char filename[0];
};

#define MODULE_DIR "/lib/modules"

static void fatal(const char *fmt, ...)
__attribute__ ((noreturn, format (printf, 1, 2)));

static void fatal(const char *fmt, ...)
{
	va_list arglist;

	fprintf(stderr, "FATAL: ");

	va_start(arglist, fmt);
	vfprintf(stderr, fmt, arglist);
	va_end(arglist);

	exit(1);
}

static void warn(const char *fmt, ...)
__attribute__ ((format (printf, 1, 2)));

static void warn(const char *fmt, ...)
{
	va_list arglist;

	fprintf(stderr, "WARNING: ");

	va_start(arglist, fmt);
	vfprintf(stderr, fmt, arglist);
	va_end(arglist);
}

static void *do_nofail(void *ptr, const char *file, int line, const char *expr)
{
	if (!ptr) {
		fatal("Memory allocation failure %s line %d: %s.\n",
		      file, line, expr);
	}
	return ptr;
}

#define NOFAIL(ptr)	do_nofail((ptr), __FILE__, __LINE__, #ptr)

static void print_usage(const char *progname)
{
	fprintf(stderr,
		"Usage: %s [--verbose|--version|--config|--remove] filename options\n",
		progname);
	exit(1);
}

static int fgetc_wrapped(FILE *file)
{
	for (;;) {
	  	int ch = fgetc(file);
		if (ch != '\\')
			return ch;
		ch = fgetc(file);
		if (ch != '\n')
			return ch;
	}
}

static char *getline_wrapped(FILE *file)
{
	int size = 1024;
	int i = 0;
	char *buf = NOFAIL(malloc(size));
	for(;;) {
		int ch = fgetc_wrapped(file);
		if (i == size) {
			size *= 2;
			buf = NOFAIL(realloc(buf, size));
		}
		if (ch < 0 || ch == '\n') {
			if (ch < 0 && i == 0) {
				free(buf);
				return NULL;
			}
			buf[i] = '\0';
			return NOFAIL(realloc(buf, i+1));
		}
		buf[i++] = ch;
	}
}

static struct module *find_module(const char *filename, struct list_head *list)
{
	struct module *i;

	list_for_each_entry(i, list, list) {
		if (strcmp(i->filename, filename) == 0)
			return i;
	}
	return NULL;
}

static void add_module(char *filename, int namelen, struct list_head *list)
{
	struct module *mod;

	/* If it's a duplicate: move it to the end, so it gets
	   inserted where it is *first* required. */
	mod = find_module(filename, list);
	if (mod)
		list_del(&mod->list);
	else {
		/* No match.  Create a new module. */
		mod = NOFAIL(malloc(sizeof(struct module) + namelen + 1));
		memcpy(mod->filename, filename, namelen);
		mod->filename[namelen] = '\0';
	}

	list_add_tail(&mod->list, list);
}

/* Compare, with _ and - equivalent. */
static int modname_equal(const char *a, const char *b, unsigned int len)
{
	unsigned int i;

	for (i = 0; i < len; i++) {
		if ((a[i] == '_' || a[i] == '-')
		    && (b[i] == '_' || b[i] == '-'))
			continue;
		if (a[i] != b[i])
			return 0;
	}
	return 1;
}

/* Returns list of modules if this is the line we want. */
static void add_modules_dep_line(char *line,
				 const char *name,
				 struct list_head *list)
{
	char *ptr;
	int len;
	char *modname;

	/* Ignore lines without : or which start with a # */
	ptr = index(line, ':');
	if (ptr == NULL || line[strspn(line, "\t ")] == '#')
		return;

	/* Is this the module we are looking for? */
	*ptr = '\0';
	if (strrchr(line, '/'))
		modname = strrchr(line, '/') + 1;
	else
		modname = line;

	len = strlen(modname);
	if (strchr(modname, '.'))
		len = strchr(modname, '.') - modname;
	if (!modname_equal(modname, name, len))
		return;

	/* Create the list. */
	add_module(line, ptr - line, list);

	ptr++;
	for(;;) {
		char *dep_start;
		ptr += strspn(ptr, " \t");
		if (*ptr == '\0')
			break;
		dep_start = ptr;
		ptr += strcspn(ptr, " \t");
		add_module(dep_start, ptr - dep_start, list);
	}
}

static void read_depends(const char *dirname,
			 const char *start_name,
			 struct list_head *list)
{
	char modules_dep_name[strlen(dirname) + sizeof("modules.dep") + 1];
	char *line;
	FILE *modules_dep;

	sprintf(modules_dep_name, "%s/%s", dirname, "modules.dep");
	modules_dep = fopen(modules_dep_name, "r");
	if (!modules_dep)
		fatal("Could not load %s: %s\n",
		      modules_dep_name, strerror(errno));

	while ((line = getline_wrapped(modules_dep)) != NULL) {
		add_modules_dep_line(line, start_name, list);
		free(line);
	}
	fclose(modules_dep);
}

/* We use error numbers in a loose translation... */
static const char *moderror(int err)
{
	switch (err) {
	case ENOEXEC:
		return "Invalid module format";
	case ENOENT:
		return "Unknown symbol in module";
	default:
		return strerror(err);
	}
}

static int
module_in_kernel(const char *modname)
{
	FILE *proc_modules;
	char *line;
	const int modname_len = strlen(modname);
	
	proc_modules = fopen("/proc/modules", "r");
	if (proc_modules == NULL)
		fatal("Error reading /proc/modules: %s\n", strerror(errno));

	while ((line = getline_wrapped(proc_modules)) != NULL) {
		if (strncmp(line, modname, modname_len) == 0 &&
		    isspace(line[modname_len])) {
			free(line);
			fclose(proc_modules);
			return 1;
		}
		free(line);
	}
	fclose(proc_modules);
	return 0;
}

static void *read_in(int fd, unsigned long len)
{
	void *ret = NOFAIL(malloc(len));
	unsigned long done = 0;
	int r;

	while (done < len) {
		r = read(fd, ret + done, len - done);
		if (r <= 0) {
			free(ret);
			return NULL;
		}
		done += r;
	}
	return ret;
}

static void rename_module32(struct module *module,
			    void *mod,
			    unsigned long len,
			    const char *newname)
{
	Elf32_Ehdr *hdr = mod;
	Elf32_Shdr *sechdrs = mod + hdr->e_shoff;
	const char *secnames = mod + sechdrs[hdr->e_shstrndx].sh_offset;
	unsigned int i;

	/* Replace the .gnu.linkonce.modname section */
	for (i = 1; i < hdr->e_shnum; i++) {
		if (strcmp(secnames+sechdrs[i].sh_name,
			   ".gnu.linkonce.modname") == 0) {
			strncpy(mod + sechdrs[i].sh_offset, newname,
				sechdrs[i].sh_size - 1);
			return;
		}
	}
	warn("Could not find module name to change in %s\n", module->filename);
}

static void rename_module64(struct module *module,
			    void *mod,
			    unsigned long len,
			    const char *newname)
{
	Elf64_Ehdr *hdr = mod;
	Elf64_Shdr *sechdrs = mod + hdr->e_shoff;
	const char *secnames = mod + sechdrs[hdr->e_shstrndx].sh_offset;
	unsigned int i;

	/* Replace the .gnu.linkonce.modname section */
	for (i = 1; i < hdr->e_shnum; i++) {
		if (strcmp(secnames+sechdrs[i].sh_name,
			   ".gnu.linkonce.modname") == 0) {
			strncpy(mod + sechdrs[i].sh_offset, newname,
				sechdrs[i].sh_size - 1);
			return;
		}
	}
	warn("Could not find module name to change in %s\n", module->filename);
}

static void rename_module(struct module *module,
			  void *mod,
			  unsigned long len,
			  const char *newname)
{
	/* "\177ELF" <byte> where byte = 001 for 32-bit, 002 for 64 */
	char *ident = mod;

	if (memcmp(mod, ELFMAG, SELFMAG) != 0) {
		warn("Unknown module format in %s: not changing name\n",
		     module->filename);
		return;
	}

	switch (ident[EI_CLASS]) {
	case ELFCLASS32:
		rename_module32(module, mod, len, newname);
		break;
	case ELFCLASS64:
		rename_module64(module, mod, len, newname);
		break;
	default:
		warn("Module %s has elf unknown identifier %i\n",
		     module->filename, ident[EI_CLASS]);
	}
}

struct module_options
{
	struct module_options *next;
	char *modulename;
	char *options;
};

struct module_installcmd
{
	struct module_installcmd *next;
	char *modulename;
	char *installcmd;
};

/* Link in a new option line from the config file. */
static struct module_options *
add_options(const char *modname,
	    const char *option,
	    struct module_options *options)
{
	struct module_options *new;

	new = NOFAIL(malloc(sizeof(*new)));
	new->modulename = NOFAIL(strdup(modname));
	new->options = NOFAIL(strdup(option));
	new->next = options;
	return new;
}

/* Link in a new install line from the config file. */
static struct module_installcmd *
add_installcmd(const char *modname,
	       const char *installcmd,
	       struct module_installcmd *installcmds)
{
	struct module_installcmd *new;

	new = NOFAIL(malloc(sizeof(*new)));
	new->modulename = NOFAIL(strdup(modname));
	new->installcmd = NOFAIL(strdup(installcmd));
	new->next = installcmds;
	return new;
}

/* Find install commands if any. */
static const char *find_installcmd(const char *modname,
				   const struct module_installcmd *installcmds)
{
	while (installcmds) {
		if (strcmp(installcmds->modulename, modname) == 0)
			return installcmds->installcmd;
		installcmds = installcmds->next;
	}
	return NULL;
}

/* Add to options */
static char *add_extra_options(const char *modname,
			       const char *base_options,
			       const struct module_options *options)
{
	char *opts = NOFAIL(strdup(base_options));
	
	while (options) {
		if (strcmp(options->modulename, modname) == 0) {
			opts = NOFAIL(realloc(opts,
					      strlen(opts) + 1
					      + strlen(options->options) + 1));
			strcat(opts, " ");
			strcat(opts, options->options);
		}
		options = options->next;
	}
	return opts;
}

/* Do an install command. */
static void do_installcmd(const char *modname,
			  const char *installcmd,
			  int verbose, int dry_run, int dont_fail)
{
	int ret;

	if (verbose)
		printf("install %s: %s\n", modname, installcmd);
	if (dry_run)
		return;

	ret = system(installcmd);
	if (ret == -1 || WEXITSTATUS(ret)) {
		if (dont_fail)
			fatal("Error running install command for %s\n",
			      modname);
		warn("Error running install command for %s\n",modname);
	}
}

static void filename2modname(char *modname, const char *filename)
{
	const char *afterslash;
	unsigned int i;

	afterslash = strrchr(filename, '/');
	if (!afterslash)
		afterslash = filename;
	else
		afterslash++;

	/* Convert to underscores, stop at first . */
	for (i = 0; afterslash[i] && afterslash[i] != '.'; i++) {
		if (afterslash[i] == '-')
			modname[i] = '_';
		else
			modname[i] = afterslash[i];
	}
	modname[i] = '\0';
}

/* Actually do the insert. */
static void insmod(struct list_head *list,
		   const char *base_options,
		   const char *newname,
		   int dont_fail,
		   int dry_run,
		   int verbose,
		   const struct module_options *options,
		   const struct module_installcmd *installcmds)
{
	int fd, ret;
	struct stat st;
	void *map;
	const char *installcmd;
	char *optstring;
	struct module *mod = list_entry(list->next, struct module, list);
	char modname[strlen(mod->filename) + 1];

	/* Take us off the list. */
	list_del(&mod->list);

	/* Do things we depend on first. */
	if (!list_empty(list))
		insmod(list, "", NULL, 0, dry_run, verbose, options,
		       installcmds);

	filename2modname(modname, mod->filename);
	if (module_in_kernel(newname ?: modname)) {
		if (dont_fail)
			fatal("Module %s already in kernel.\n",
			      newname ?: modname);
		return;
	}

	/* Open and suck in module (don't mmap, we might want to mangle) */
	fd = open(mod->filename, O_RDONLY);
	if (fd < 0) {
		if (dont_fail)
			fatal("Could not open `%s': %s\n",
			      mod->filename, strerror(errno));
		warn("Could not open `%s': %s\n",
		     mod->filename, strerror(errno));
		return;
	}

	fstat(fd, &st);
	map = read_in(fd, st.st_size);
	if (!map) {
		if (dont_fail)
			fatal("Can't read `%s': %s\n",
			      mod->filename, strerror(errno));
		warn("Can't read `%s': %s\n", mod->filename, strerror(errno));
		goto out_fd;
	}

	/* Rename it? */
	if (newname)
		rename_module(mod, map, st.st_size, newname);

	/* Did config file override installcmd or add options? */
	installcmd = find_installcmd(modname, installcmds);
	optstring = add_extra_options(modname, base_options, options);

	if (installcmd) {
		do_installcmd(modname, installcmd,
			      verbose, dry_run, dont_fail);
		goto out;
	}

	if (verbose)
		printf("insmod %s %s\n", mod->filename, optstring);

	if (dry_run)
		goto out;

	ret = syscall(__NR_init_module, map, st.st_size, optstring);
	if (ret != 0) {
		if (dont_fail)
			fatal("Error inserting %s (%s): %s\n",
			      modname, mod->filename, moderror(errno));
		warn("Error inserting %s (%s): %s\n",
		     modname, mod->filename, moderror(errno));
	}
 out:
	free(map);
	free(optstring);
 out_fd:
	close(fd);
}

/* Do recursive removal. */
static void rmmod(struct list_head *list,
		  int dont_fail,
		  int dry_run,
		  int verbose)
{
	struct module *mod = list_entry(list->next, struct module, list);
	char modname[strlen(mod->filename) + 1];

	/* Take first one off the list. */
	list_del(&mod->list);

	filename2modname(modname, mod->filename);
	if (!module_in_kernel(modname)) {
		if (dont_fail)
			fatal("Module %s is not in kernel.\n", modname);
		goto remove_rest;
	}

	if (verbose)
		printf("rmmod %s\n", mod->filename);

	if (dry_run)
		goto remove_rest;

	if (syscall(__NR_delete_module, modname, O_EXCL) != 0) {
		if (dont_fail)
			fatal("Error removing %s (%s): %s\n",
			      modname, mod->filename, moderror(errno));
		warn("Error removing %s (%s): %s\n",
		     modname, mod->filename, moderror(errno));
	}

 remove_rest:
	/* Now do things we depend. */
	if (!list_empty(list))
		rmmod(list, 0, dry_run, verbose);
}

static char *basename(char *name)
{
	char *base = strrchr(name, '/');
	if (base) return base + 1;
	return name;
}

static void underscores(char *string)
{
	unsigned int i;
	for (i = 0; string[i]; i++)
		if (string[i] == '-')
			string[i] = '_';
}

static int do_modprobe(const char *argv0, const char *modname)
{
	int ret;

	ret = fork();
	if (ret < 0)
		fatal("Failed to fork to run modprobe %s: %s\n",
		      modname, strerror(errno));
	if (ret > 0) {
		int status;
		waitpid(ret, &status, 0);
		if (WIFEXITED(status))
			ret = (WEXITSTATUS(status) != 0);
		else ret = 1;
	} else {
		execlp(argv0, argv0, modname, NULL);
		fatal("Failed to run modprobe %s: %s\n",
		      modname, strerror(errno));
	}

	return ret;
}

/* Final dir of path must match subpath. */
static int type_matches(const char *path, const char *subpath)
{
	char *prev, *slash = strrchr(path, '/');
	if (!slash)
		return 0;

	prev = slash - (strlen(subpath)+1);
	if (prev < path)
		return 0;

	if (prev[0] == '/' && memcmp(prev+1, subpath, strlen(subpath)) == 0)
		return 1;
	return 0;
}

static int do_wildcard(const char *argv0,
		       const char *dirname,
		       int list_only,
		       const char *type,
		       int all,
		       int dry_run,
		       int verbose,
		       char *wildcard)
{
	char modules_dep_name[strlen(dirname) + sizeof("modules.dep") + 1];
	char *line;
	FILE *modules_dep;
	int ret;

	if (list_only) /* List is always "OK" even if nothing matches */
		ret = 0;
	else /* Otherwise, only success of one succeeds. */
		ret = 1;

	/* Canonicalize wildcard. */
	underscores(wildcard);

	sprintf(modules_dep_name, "%s/%s", dirname, "modules.dep");
	modules_dep = fopen(modules_dep_name, "r");
	if (!modules_dep)
		fatal("Could not load %s: %s\n",
		      modules_dep_name, strerror(errno));

	while ((line = getline_wrapped(modules_dep)) != NULL) {
		char *ptr;
		char modpath[strlen(line)+1];

		/* Ignore lines without : or which start with a # */
		ptr = strchr(line, ':');
		if (ptr == NULL || line[strspn(line, "\t ")] == '#')
			goto next;
		*ptr = '\0';
		strcpy(modpath, line);

		/* "type" must match final directory part. */
		if (type && !type_matches(modpath, type))
			goto next;

		/* Strip extension */
		ptr = strrchr(line, '.');
		if (ptr)
			*ptr = '\0';

		ptr = basename(line);
		underscores(ptr);

		if (fnmatch(wildcard, ptr, 0) == 0) {
			if (list_only)
				printf("%s\n", modpath);
			else {
				if (verbose)
					printf("modprobe %s\n", ptr);
				if (!dry_run)
					ret &= do_modprobe(argv0, ptr);
				/* They might not want us to load all. */
				if (ret == 0 && !all) {
					free(line);
					return 0;
				}
			}
		}
	next:
		free(line);
	}

	return ret;
}

static char *strsep_skipspace(char **string, char *delim)
{
	*string += strspn(*string, delim);
	return strsep(string, delim);
}

/* FIXME: Maybe should be extended to "alias a b [and|or c]...".
   modules.dep must also contain aliases extracted from modules. --RR */

/* Simple format, ignore lines starting with #, one command per line.
   Returns NULL or resolved alias. */
static char *read_config(const char *filename,
			 int mustload,
			 char *name,
			 int dump_only,
			 struct module_options **options,
			 struct module_installcmd **installcmds)
{
	FILE *cfile;
	char *line;
	char *result = NULL;

	cfile = fopen(filename, "r");
	if (!cfile) {
		if (mustload)
			fatal("Failed to open config file %s: %s\n",
			      filename, strerror(errno));
		return NOFAIL(strdup(name));
	}

	while ((line = getline_wrapped(cfile)) != NULL) {
		char *ptr = line;
		char *cmd, *modname;

		if (dump_only)
			printf("%s\n", line);

		cmd = strsep_skipspace(&ptr, "\t ");
		if (cmd == NULL)
			continue;

		if (strcmp(cmd, "alias") == 0) {
			char *fakename = strsep_skipspace(&ptr, "\t ");
			char *realname = strsep_skipspace(&ptr, "\t ");

			if (fakename && realname && !strcmp(fakename, name))
				result = NOFAIL(strdup(realname));
		} else if (strcmp(cmd, "include") == 0) {
			filename = strsep(&ptr, "\t ");
			result = read_config(filename, 0, name, dump_only,
					     options, installcmds);
		} else if (strcmp(cmd, "options") == 0) {
			modname = strsep(&ptr, "\t ");
			*options = add_options(modname, ptr, *options);
		} else if (strcmp(cmd, "install") == 0) {
			modname = strsep(&ptr, "\t ");
			*installcmds =add_installcmd(modname,ptr,*installcmds);
		}
		free(line);
	}
	fclose(cfile);
	return result;
}

static struct option options[] = { { "verbose", 0, NULL, 'v' },
				   { "version", 0, NULL, 'V' },
				   { "config", 1, NULL, 'C' },
				   { "name", 1, NULL, 'o' },
				   { "remove", 0, NULL, 'r' },
				   { "showconfig", 0, NULL, 'c' },
				   { "autoclean", 0, NULL, 'k' },
				   { "quiet", 0, NULL, 'q' },
				   { "show", 0, NULL, 'n' },
				   { "syslog", 0, NULL, 's' },
				   { "type", 1, NULL, 't' },
				   { "list", 0, NULL, 'l' },
				   { NULL, 0, NULL, 0 } };

#define DEFAULT_CONFIG "/etc/modprobe.conf"

int main(int argc, char *argv[])
{
	struct utsname buf;
	int opt;
	int dump_only = 0;
	int dry_run = 0;
	int remove = 0;
	int verbose = 0;
	int list_only = 0;
	int all = 0;
	char *type = NULL;
	const char *config = NULL, *installcmd;
	char *dirname, *optstring = NOFAIL(strdup(""));
	char *modname, *newname = NULL;
	struct module_installcmd *installcmds = NULL;
	struct module_options *modoptions = NULL;
	LIST_HEAD(list);

	try_old_version("modprobe", argv);

	while ((opt = getopt_long(argc, argv, "vVC:o:rknqsclt:a", options, NULL)) != -1){
		switch (opt) {
		case 'v':
			verbose = 1;
			break;
		case 'V':
			puts(PACKAGE " version " VERSION);
			exit(0);
		case 'C':
			config = optarg;
			break;
		case 'o':
			newname = optarg;
			break;
		case 'r':
			remove = 1;
			break;
		case 'c':
			dump_only = 1;
			break;
		case 't':
			type = optarg;
			break;
		case 'l':
			list_only = 1;
			break;
		case 'a':
			all = 1;
			break;
		case 'k':
			/* FIXME: This should actually do something */
			break;
		case 'n':
			dry_run = 1;
			break;
		case 'q':
			/* FIXME: This should actually do something */
			break;
		case 's':
			/* FIXME: This should actually do something */
			break;
		default:
			print_usage(argv[0]);
		}
	}

	if (argc < optind + 1 && !dump_only && !list_only)
		print_usage(argv[0]);

	uname(&buf);
	dirname = NOFAIL(malloc(strlen(buf.release) + sizeof(MODULE_DIR) + 1));
	sprintf(dirname, "%s/%s", MODULE_DIR, buf.release);

	/* Old-style -t xxx wildcard?  Or -l, or -a. */
	if (all || type || list_only) {
		/* fprintf(stderr, "man find\n"); return 1; */
		return do_wildcard(argv[0], dirname, list_only, type, all,
				   dry_run, verbose, argv[optind] ?: "*");
	}

	/* Rest is module options */
	for (opt = optind + 1; opt < argc; opt++) {
		optstring = NOFAIL(realloc(optstring,
					   strlen(optstring) + 2
					   + strlen(argv[opt]) + 2));
		/* Spaces handled by "" pairs, but no way of escaping
                   quotes */
		if (strchr(argv[opt], ' '))
			strcat(optstring, "\"");
		strcat(optstring, argv[opt]);
		if (strchr(argv[opt], ' '))
			strcat(optstring, "\"");
		strcat(optstring, " ");
	}

	/* Returns the resolved alias, options */
	modname = read_config(config ?: DEFAULT_CONFIG,
			      config ? 1 : 0,
			      dump_only ? "" : argv[optind],
			      dump_only,
			      &modoptions, &installcmds);
	if (dump_only)
		exit(0);

	if (modname == NULL)
		modname = NOFAIL(strdup(argv[optind]));

	/* The dependencies have to be real modules, but the first one
           can be completely bogus (as long as there's an installcmd). */
	read_depends(dirname, modname, &list);
	installcmd = find_installcmd(modname, installcmds);
	if (remove) {
		/* FIXME: Implement "remove" command --RR */
		if (list_empty(&list))
			fatal("Module %s not found.\n", modname);
		rmmod(&list, 1, dry_run, verbose);
	} else {
		if (!list_empty(&list)) {
			insmod(&list, optstring, newname, 1, dry_run, verbose,
			       modoptions, installcmds);
		} else {
			if (!installcmd)
				fatal("Module %s not found.\n", modname);
			do_installcmd(modname, installcmd, verbose, dry_run,1);
		}
	}
	free(dirname);
	free(modname);

	return 0;
}
